
/*********************************************************************
 * (C) Copyright 1999 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *********************************************************************/










/*********************************************************************
 * File: backdoors.c
 * Description: functions for cnf output
 *
 * Author: Joerg Hoffmann 2003
 *  Modified: Shane Neph 2004 and 2006
 *********************************************************************/ 




#include "vec.h"
#include "bb.h"

#include "output.h"
#include "memory.h"

#include "instantiateI.h"
#include "instantiateII.h"

#include "graph.h"

#include "cnfout.h"
#include "cg_encoding.h"
#include "solver.h"
#include "utilities.h"

#include <time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h> 
#include <pthread.h>
#include <assert.h>

int **lclause;
int *lclause_size;
int lnum_clauses;


extern const int UNSAT; //2
extern const int SAT;	//0
//Q. Lv 2008-3-31
extern int gsolution_num;
extern int gmini_costs;
extern int gmini_costs_layer;
extern float  gmini_costs_time;
extern int gcurr_layer;
extern double global_time;
extern struct tms gstart,gend;

pthread_mutex_t layer_lock = PTHREAD_MUTEX_INITIALIZER;

float encoding_time = 0, solving_time = 0;

int do_cnf_output(solver* minisat,int create)
{
	float time_slice;
	gcmd_line.cnfout = 4;		
	times(&start);
	
	int result;
    if ( gcmd_line.cnfout == 1 ) {
    	
		result = print_action_based_encoding(minisat, minisat->stats.layer, create );
    
    }else
    if ( gcmd_line.cnfout == 4 ) {
	
		result = print_thin_gp_based_encoding(minisat, minisat->stats.layer, create );
    
    }else{
	
	    printf("\n\nEXIT: cnf out %d not implemented\n\n", gcmd_line.cnfout);
		exit(1);
    }

	times(&end);
	TIME( encoding_time );
    return result;
}



extern struct DisAction* distance_action;
int *F = NULL, *A = NULL;   
    
/* action-based
 */
int print_action_based_encoding(solver* minisat, int layer, int create )
{
	check_memory_usage();
    FILE *CNF, *VARFILE;

    int i, j, t, k, ft, op, prevcl, l, result=0;
    int tmpNumber;

    int *lop_pos1, *lop_pos2;
    
    int nGclauses = 0, nAclauses = 0, nEclauses = 0;

    lclause = ( int ** ) calloc( MAX_CLAUSES, sizeof( int * ) );
    lclause_size = ( int * ) calloc( MAX_CLAUSES, sizeof( int ) );
    lop_pos1 = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );  
    lop_pos2 = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );      
    lcode = ( int ** ) calloc( layer, sizeof( int * ) );
    lop_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    ltime_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    for ( i = 0; i < layer; i++ ) {
		lcode[i] = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
		for ( j = 0; j < gnum_op_conn; j++ )
		    lcode[i][j] = -1;
    }
    
	for(i=0;i<MAX_CNF_VARS;i++) lop_pos1[i]=-1;

    F  = ( int * ) calloc( gnum_ft_conn, sizeof( int ) );
    for ( i = 0; i < gnum_ft_conn; i++ )
		F[i] = -1;
		
    /* no-ops are included in the gop_conn!
     */
    A  = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    for ( i = 0; i < gnum_op_conn; i++ )
		A[i] = -1;

    /* layer 0 */
    for ( i = 0; i < ginitial_state.num_F; i++ )
		F[ginitial_state.F[i]] = 0;
	
    printf("\ncreating action-based encoding:");

    /* prec constraints, + code assignment.
     */
    lnum_clauses = 0;
    code = 1; /* variables numbered 1 .. n */
    printf("\n        building rplan graph layers, up to layer [%d];", layer);
    for ( t = 0; t < layer; t++ ) {	
	for ( i = 0; i < gnum_op_conn; i++ ) {
	    if ( A[i] == -1 ) {
			for ( j = 0; j < gop_conn[i].num_P; j++ )
			    if ( F[gop_conn[i].P[j]] == -1 || F[gop_conn[i].P[j]] > t ) 
			    	break;
			if ( j < gop_conn[i].num_P ) continue;
			/* now i is an op that comes in here.
			 */
			A[i] = t;
			/* mark its add effects.
			 */
			for ( j = 0; j < gop_conn[i].num_A; j++ ) {
			    ft =  gop_conn[i].A[j];
			    if ( F[ft] != -1 ) continue;
			    F[ft] = t+1;
			}
	    }

	    /* insert prec clauses, if at t > 0.
	     */
	    if ( t == 0 ) {
			continue;
	    }
	    for ( j = 0; j < gop_conn[i].num_P; j++ ) {
			ft =  gop_conn[i].P[j];
	
			if ( lnum_clauses == MAX_CLAUSES ) {
			    printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
			    exit( 1 );
			}
			lclause[lnum_clauses] = ( int * ) calloc( gft_conn[ft].num_A+1, sizeof( int ) );
			/* here the op i is used at t, for the 1st time. assign the code.
			 */
			if ( j == 0 ) {
			    if ( lcode[t][i] != -1 ) {
					printf("\n\nEXIT: op code %d at %d already assigned??\n\n", i, t);
					exit( 1 );
			    }
			    lcode[t][i] = code++;
			    if ( code == MAX_CNF_VARS + 1 ) {
					printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					exit( 1 );
			    }
			    lop_to_code[code-1] = i;
			    ltime_to_code[code-1] = t;
			}
	
			lclause[lnum_clauses][0] = (-1) * lcode[t][i];
			lclause_size[lnum_clauses] = 1;
			for ( k = 0; k < gft_conn[ft].num_A; k++ ) {
			    op = gft_conn[ft].A[k];
			    if ( A[op] == -1 || A[op] > t-1 )
					continue;
			    if ( lcode[t-1][op] == -1 ) {
					lcode[t-1][op] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = op;
					ltime_to_code[code-1] = t-1;
			    }
			    lclause[lnum_clauses][lclause_size[lnum_clauses]++] = lcode[t-1][op];
			}
			if ( lclause_size[lnum_clauses] == 1 ) {
				lnum_clauses++;
			    printf("\n\nEXIT: no achiever in at t>0??\n\n");
			    //exit( 1 );
			    result = UNSAT;
			    goto freeandquit;
			}
			nAclauses++;
			lnum_clauses++;
	    } /* pres of i-op */
	} /* i -- op */
    } /* t */


    /* goal constraints
     */
    printf("\n        goal constraints;");
    for ( i = 0; i < ggoal_state.num_F; i++ ) {
		ft = ggoal_state.F[i];
	
		if ( lnum_clauses == MAX_CLAUSES ) {
		    printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
		    exit( 1 );
		}
		lclause[lnum_clauses] = ( int * ) calloc( gft_conn[ft].num_A, sizeof( int ) );
		lclause_size[lnum_clauses] = 0;
		for ( k = 0; k < gft_conn[ft].num_A; k++ ) {
		    op = gft_conn[ft].A[k];
		    if ( A[op] == -1 || A[op] > layer-1 )
				continue;
		    if ( lcode[layer-1][op] == -1 ) {
				lcode[layer-1][op] = code++;
				if ( code == MAX_CNF_VARS + 1 ) {
				    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
				    exit( 1 );
				}
				lop_to_code[code-1] = op;
				ltime_to_code[code-1] = layer-1;
		    }
		    lclause[lnum_clauses][lclause_size[lnum_clauses]++] = lcode[layer-1][op];
		}
		if ( lclause_size[lnum_clauses] == 0 ) {
		    printf("\n\nno achiever in for goal?? deadline too low!#001\n\n");
		    lnum_clauses++;
		    //exit(UNSAT);
		    result = UNSAT;
		    goto freeandquit;
		}
		lnum_clauses++;
		nGclauses++;
    } /* goals */

 
    /* exclusion constraints. implementation a bit smart, to avoid
     * time-consuming all-op-pairs-check.
     */
	printf("\n        building exclusion constraints --> up to layer %d...", layer);fflush(stdout);
    for ( t = 0; t < layer; t++ ) {
	for ( i = 0; i < gnum_op_conn; i++ ) {
	    if ( A[i] == -1 || A[i] > t ) continue;
	    prevcl = lnum_clauses;
	    for ( j = 0; j < gop_conn[i].num_D; j++ ) {
			ft = gop_conn[i].D[j];
			for ( k = 0; k < gft_conn[ft].num_P; k++ ) {
			    op = gft_conn[ft].P[k];
			    if ( op <= i ) continue; /* only in one of the two dirs we'll get */
			    if ( A[op] == -1 || A[op] > t ) continue;
			    /* did we make op excl of i already?
			     */
			    if ( lcode[t][op] != -1 ) {
					for ( l = prevcl; l < lnum_clauses; l++ ) {
					    if ( lclause[l][1] == (-1) * lcode[t][op] ) break;
					}
					if ( l < lnum_clauses ) continue; /* yes. */
			    }
			    /* record the clause.
			     */
			    if ( lnum_clauses == MAX_CLAUSES ) {
					printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
					exit( 1 );
			    }
			    lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
			    if ( lcode[t][i] == -1 ) {
					lcode[t][i] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = i;
					ltime_to_code[code-1] = t;
			    }
				lop_pos1[lcode[t][i]]=prevcl;
			    lclause[lnum_clauses][0] = (-1) * lcode[t][i];
			    if ( lcode[t][op] == -1 ) {
					lcode[t][op] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = op;
					ltime_to_code[code-1] = t;
			    }
			    lclause[lnum_clauses][1] = (-1) * lcode[t][op];
			    lclause_size[lnum_clauses] = 2;
				lop_pos2[lcode[t][i]]=lnum_clauses;
			    lnum_clauses++;
			    nEclauses++;
			} /* k: ops that ft is pre of */
			for ( k = 0; k < gft_conn[ft].num_A; k++ ) {
			    op = gft_conn[ft].A[k];
			    if ( op <= i ) continue; /* only in one of the two dirs we'll get */
			    if ( A[op] == -1 || A[op] > t ) continue;
			    /* did we make op excl of i already?
			     */
			    if ( lcode[t][op] != -1 ) {
					for ( l = prevcl; l < lnum_clauses; l++ ) {
					    if ( lclause[l][1] == (-1) * lcode[t][op] ) break;
					}
					if ( l < lnum_clauses ) continue; /* yes. */
			    }
			    /* record the clause.
			     */
			    if ( lnum_clauses == MAX_CLAUSES ) {
					printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
					exit( 1 );
			    }
			    lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
			    if ( lcode[t][i] == -1 ) {
					lcode[t][i] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = i;
					ltime_to_code[code-1] = t;
			    }
				lop_pos1[lcode[t][i]]=prevcl;			    
			    lclause[lnum_clauses][0] = (-1) * lcode[t][i];
			    if ( lcode[t][op] == -1 ) {
					lcode[t][op] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = op;
					ltime_to_code[code-1] = t;
			    }
			    lclause[lnum_clauses][1] = (-1) * lcode[t][op];
			    lclause_size[lnum_clauses] = 2;
				lop_pos2[lcode[t][i]]=lnum_clauses;
			    lnum_clauses++;
			    nEclauses++;
			} /* k: ops that ft is added by */
	    } /* j: fts that i dels */


	    for ( j = 0; j < gop_conn[i].num_P; j++ ) {
			ft = gop_conn[i].P[j];
			for ( k = 0; k < gft_conn[ft].num_D; k++ ) {
			    op = gft_conn[ft].D[k];
			    if ( op <= i ) continue; /* only in one of the two dirs we'll get */
			    if ( A[op] == -1 || A[op] > t ) continue;
			    /* did we make op excl of i already?
			     */
			    if ( lcode[t][op] != -1 ) {
					for ( l = prevcl; l < lnum_clauses; l++ ) {
					    if ( lclause[l][1] == (-1) * lcode[t][op] ) break;
					}
					if ( l < lnum_clauses ) continue; /* yes. */
			    }
			    /* record the clause.
			     */
			    if ( lnum_clauses == MAX_CLAUSES ) {
				printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
				exit( 1 );
			    }
			    lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
			    if ( lcode[t][i] == -1 ) {
					lcode[t][i] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = i;
					ltime_to_code[code-1] = t;
			    }
				lop_pos1[lcode[t][i]]=prevcl;			    
			    lclause[lnum_clauses][0] = (-1) * lcode[t][i];
			    if ( lcode[t][op] == -1 ) {
					lcode[t][op] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = op;
					ltime_to_code[code-1] = t;
			    }
			    lclause[lnum_clauses][1] = (-1) * lcode[t][op];
			    lclause_size[lnum_clauses] = 2;
				lop_pos2[lcode[t][i]]=lnum_clauses;			    
			    lnum_clauses++;
			    nEclauses++;
			} /* k: ops that ft is del by */
	    } /* j: fts that i has as prec */

	    for ( j = 0; j < gop_conn[i].num_A; j++ ) {
			ft = gop_conn[i].A[j];
			for ( k = 0; k < gft_conn[ft].num_D; k++ ) {
			    op = gft_conn[ft].D[k];
			    if ( op <= i ) continue; /* only in one of the two dirs we'll get */
			    if ( A[op] == -1 || A[op] > t ) continue;
			    /* did we make op excl of i already?
			     */
			    if ( lcode[t][op] != -1 ) {
					for ( l = prevcl; l < lnum_clauses; l++ ) {
					    if ( lclause[l][1] == (-1) * lcode[t][op] ) break;
					}
					if ( l < lnum_clauses ) continue; /* yes. */
			    }
			    /* record the clause.
			     */
			    if ( lnum_clauses == MAX_CLAUSES ) {
					printf("\n\nEXIT: too many clauses. increase MAX_CLAUSES.\n\n");
					exit( 1 );
			    }
			    lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
			    if ( lcode[t][i] == -1 ) {
					lcode[t][i] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = i;
					ltime_to_code[code-1] = t;
			    }
				lop_pos1[lcode[t][i]]=prevcl;		    
			    lclause[lnum_clauses][0] = (-1) * lcode[t][i];
			    if ( lcode[t][op] == -1 ) {
					lcode[t][op] = code++;
					if ( code == MAX_CNF_VARS + 1 ) {
					    printf("\n\nEXIT: too many cnf vars. increase MAX_CNF_VARS.\n\n");
					    exit( 1 );
					}
					lop_to_code[code-1] = op;
					ltime_to_code[code-1] = t;
			    }
			    lclause[lnum_clauses][1] = (-1) * lcode[t][op];
			    lclause_size[lnum_clauses] = 2;
				lop_pos2[lcode[t][i]]=lnum_clauses;		    
			    lnum_clauses++;
			    nEclauses++;
			} /* k: ops that ft is del by */
	    } /* j: fts that i adds */
	} /* i: ops at t */
    } /* t */


    /* that's it. print CNF file.
     */
    if ( gcmd_line.debug ) {
		if ( (VARFILE = fopen( gcmd_line.varFileName, "w" )) == NULL ) {
		    printf("\n\nEXIT: can't open VARFILE file.\n\n");
		    exit( 1 );
		}
		fprintf(VARFILE, "\n\nc DECISION LAYER %d, ACTION-BASED ENCODING.", layer);
		fprintf(VARFILE, "\n\nc VARS:");
	
	        
		for ( i = 1; i < code; i++ ) {
			fprintf(VARFILE, "\nc var %2d action (", i);
	  	    print_op_nameToFile( lop_to_code[i], VARFILE );
		    fprintf(VARFILE, ") %d", ltime_to_code[i] );
		} 
	
/*
		fprintf(VARFILE, "\n\nCLAUSES:");
		for ( i = 0; i < lnum_clauses; i++ ) {
	   	    for ( j = 0; j < lclause_size[i]; j++ ) {
			if ( lclause[i][j] > 0 ) {
	                    fprintf(VARFILE, "\n%6d - ", i);
			    fprintf(VARFILE, "(");
			    print_op_nameToFile( lop_to_code[lclause[i][j]], VARFILE );
			    fprintf(VARFILE, " at %d) VAR = %d", ltime_to_code[lclause[i][j]], lclause[i][j] );
		        } else {
	                    fprintf(VARFILE, "\n%6d - ", i);
			    fprintf(VARFILE, "(NOT ");
			    print_op_nameToFile( lop_to_code[(-1)*lclause[i][j]], VARFILE );
			    fprintf(VARFILE, " at %d) VAR = %d", ltime_to_code[(-1)*lclause[i][j]], lclause[i][j] );
			}
		    }
		}
		fprintf(VARFILE, "0\n");
*/
        fclose(VARFILE);
    }

    printf("\n\nDECISION LAYER %d, WRITING ACTION-BASED ENCODING WITH %d VARS, %d CLAUSES.", 
	   layer, code-1, lnum_clauses);
    printf("\n%d G clauses, %d A clauses, %d E clauses.", 
	   nGclauses, nAclauses, nEclauses);
    
	
	if ( create == 0 ) {

		printf("c CNF file planning task -p %s, -o %s, -f %s, bound %d, action based encoding\n",
			gcmd_line.path, gcmd_line.ops_file_name, gcmd_line.fct_file_name, layer);
      
		tmpNumber = lnum_clauses;
		if ( gcmd_line.binary_clause_only != 0 ) {
			for ( i = 0; i < tmpNumber; ++i ) {
				if ( lclause_size[i] > 2 ) --lnum_clauses;
			}
      	}
		printf("In building up solver: p cnf %d %d\n", code-1, lnum_clauses);
      
		veci tmp_lits;
    	veci_new(&tmp_lits);
    	
		for ( i = 0; i <tmpNumber ; i++ ) {
	        lit* begin;
			veci_resize(&tmp_lits,0);
		  	
		  	for ( j = 0; j < lclause_size[i]; j++ ) {
		  		int var;
				var = abs(lclause[i][j])-1;
		    	/*fprintf(CNF, "%d ", lclause[i][j]);*/
				veci_push(&tmp_lits, (lclause[i][j] > 0 ? toLit(var) : lit_neg(toLit(var))));
			}
			
			begin = veci_begin(&tmp_lits);
	    	if (!solver_addclause( minisat, begin, begin+veci_size(&tmp_lits))){
	        	//veci_delete(&tmp_lits);
	        	//printf("CRITICAL ERROR! CANNOT ADD CLAUSE ANY MORE.\n");
	        	//exit(-1);
	        }
			free(lclause[i]);lclause[i]=NULL;
  		}
		free(lclause);
		lclause = NULL;
		veci_delete(&tmp_lits);
    }
    
  	freeandquit:
  		if(lop_pos1) free(lop_pos1);
  		if(lop_pos2) free(lop_pos2);
  		
    return result;
}



/*
 * 4th encoding
 */
int  initilized = 0;
const char CODE_TYPE_FT = 0;
const char CODE_TYPE_OP = 1;

int mutex_count=0;
int long_in_mutex_count = 0;
int londex_count = 0;
int londex_pair_count = 0;

int print_thin_gp_based_encoding(solver* minisat, int layer, int create )
{
    FILE *CNF, *VARFILE;

    int i, j, t, ft, op, code,k,l;
    IntList *tmp, *i1, *i2;
	
	int result = SAT;

    lclause = ( int ** ) calloc( MAX_CLAUSES, sizeof( int * ) );
    lclause_size = ( int * ) calloc( MAX_CLAUSES, sizeof( int ) );
    lcode = ( int ** ) calloc( layer, sizeof( int * ) );
    lop_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    ltime_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    for ( i = 0; i < layer; i++ ) {
		lcode[i] = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
		for ( j = 0; j < gnum_op_conn; j++ ) {
		    lcode[i][j] = -1;
		}
    }
    fcode = ( int ** ) calloc( layer + 1, sizeof( int * ) );
    fop_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    ftime_to_code = ( int * ) calloc( MAX_CNF_VARS, sizeof( int ) );
    for ( i = 0; i <= layer; i++ ) {
		fcode[i] = ( int * ) calloc( gnum_ft_conn, sizeof( int ) );
		for ( j = 0; j < gnum_ft_conn; j++ ) {
		    fcode[i][j] = -1;
		}
    }
	
	code_type = malloc( sizeof(char) * (gnum_op_conn + gnum_ft_conn + 1) * layer);
	code_cost = (int*)calloc( (gnum_op_conn + gnum_ft_conn + 1) * layer, sizeof( int ) );	
    /* build the graph up to <layer>
     */
     if( !initilized ){
     	initilized = 1;
	    for ( i = 0; i < gnum_op_conn; i++ ) {
			tmp = new_IntList( i );
			if ( gout_ops ) {
			    gout_ops->prev = tmp;
			}
			tmp->next = gout_ops;
			gout_ops = tmp;
	    }
	    for ( i = 0; i < ginitial_state.num_F; i++ ) {
			ft = ginitial_state.F[i];
			gin_ft_count++;
			gft_conn[ft].first_appearance = 0;
			gft_conn[ft].info_at[0] = new_FtLevelInfo();
			tmp = new_IntList( ft );
			tmp->next = gin_fts;
			gin_fts = tmp;
	    }
	    gin_fts_at[0] = gin_fts;
	    if ( gcmd_line.display_info ) {
			printf("\ntime: %3d, %5d facts and %7d exclusive pairs\n\n", 0, gin_ft_count, gin_ft_exclusion_count);
			fflush(stdout);
	    }
     }
     
	build_graph_evolution_step();
	if( gcmd_line.startLevel!=1 && layer < gcmd_line.startLevel ){
		result = UNSAT;
		goto freeandquit;
	}
#if(DEBUG)
    printf("creating thin gp-based encoding...\n");fflush(stdout);
#endif
    /* prec constraints, + code assignment.
     */
    lnum_clauses = 0;
    code = 1; /* variables numbered 1 .. n */
#if(DEBUG)
	printf("plan graph layers, at layer %d...",layer);
#endif
	for ( t = 0; t <= layer; t++ ) {
#if(DEBUG)
		printf("[%d]", t);fflush(stdout);
#endif
        for ( i = 0; i < gnum_ft_conn; i++ ) {
            if ( gft_conn[i].info_at[t] == NULL /* equiv to not in here */ ) {
                continue;
            }
            /* insert adder clause, if at t > 0. else, say it has to be true,
             * if it's an initial fact.
             */
            if ( t == 0 ) {
                for ( j = 0; j < ginitial_state.num_F; j++ ) {
                    if ( ginitial_state.F[j] == i ) break;
                }
                if ( j < ginitial_state.num_F ) {

                    fcode[t][i] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    fop_to_code[code-1] = i;
                    ftime_to_code[code-1] = t;
					code_type[code-1] = CODE_TYPE_FT;
					code_cost[code-1] = 0;
					if ( lnum_clauses == MAX_CLAUSES ) {
                        printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                        exit( 1 );
                    }
                    lclause[lnum_clauses] = ( int * ) calloc( 1, sizeof( int ) );
                    lclause[lnum_clauses][0] = fcode[t][i];
                    lclause_size[lnum_clauses++] = 1;
                }
                continue;
            }
/*			
			#if(DEBUG)
			printf("initial state finished...\n");fflush(stdout);
			#endif
*/			
            if ( fcode[t][i] != -1 ) {
                printf("\n\nft already coded??\n\n");
				result = UNSAT;
                goto freeandquit;
            }

            fcode[t][i] = code++;
            if ( code == MAX_CNF_VARS + 1 ) {
                printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                exit( 1 );
            }
            fop_to_code[code-1] = i;
            ftime_to_code[code-1] = t;
            code_type[code-1] = CODE_TYPE_FT;
			code_cost[code-1] = 0;
			if ( lnum_clauses == MAX_CLAUSES ) {
                printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                exit( 1 );
            }
            lclause[lnum_clauses] = ( int * ) calloc( gft_conn[i].num_A+1, sizeof( int ) );
            lclause[lnum_clauses][0] = (-1) * fcode[t][i];
            lclause_size[lnum_clauses] = 1;
            for ( j = 0; j < gft_conn[i].num_A; j++ ) {
                op = gft_conn[i].A[j];
                if ( gop_conn[op].info_at[t-1] == NULL ) {
                    continue;
                }
                if ( lcode[t-1][op] == -1 ) {

                    lcode[t-1][op] = code++;
                    if ( 0 ) {
                        printf("\npC %d at %d", op, t-1);
                    }
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    lop_to_code[code-1] = op;
                    ltime_to_code[code-1] = t-1;
                	code_type[code-1] = CODE_TYPE_OP;
                	//Q. Lv 2008-4-4
					code_cost[code-1] = gop_conn[op].cost;
				}
                lclause[lnum_clauses][lclause_size[lnum_clauses]++] = lcode[t-1][op];
            }
            if ( lclause_size[lnum_clauses] == 1 ) {
                printf("\n\nno achiever in at t>0??\n\n");
                result = UNSAT;
                goto freeandquit;
            }
            lnum_clauses++;
        } /* i-ft */

        if ( t == layer ) {
            continue;
        }

        for ( i = 0; i < gnum_op_conn; i++ ) {
            if ( gop_conn[i].info_at[t] == NULL /* equiv to not in here */ ) {
                continue;
            }
            /* insert prec clauses.
             */
			/*
			#if(DEBUG)
			printf("insert prec clauses\n");fflush(stdout);
			#endif
			*/
            for ( j = 0; j < gop_conn[i].num_P; j++ ) {
                ft =  gop_conn[i].P[j];

                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                /* here the op i is used at t, for the 1st time. assign the code.
                 */
                if ( j == 0 ) {
                    if ( lcode[t][i] != -1 ) {
                        printf("\n\nop code %d at %d already assigned??\n\n", i, t);
                    	result = UNSAT;
                    	goto freeandquit;
                    }
                    if ( 0 ) {
                        printf("\nC %d at %d", i, t);
                    }

                    lcode[t][i] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    lop_to_code[code-1] = i;
                    ltime_to_code[code-1] = t;
                    code_type[code-1] = CODE_TYPE_OP;
					//Q. Lv 2008-4-4
					code_cost[code-1] = gop_conn[i].cost;
                }

                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                if ( fcode[t][ft] == -1 ) {
                    printf("\n\nprec fact not coded??\n\n");
                    result = UNSAT;
                    goto freeandquit;
                }
                lclause[lnum_clauses][1] = fcode[t][ft];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            } /* j: pres of i-op */
        } /* i -- op */
    } /* t */


    /* goal constraints
     */
#if(DEBUG)
	printf("\ngoal constraints...\n");fflush(stdout);
#endif
    for ( i = 0; i < ggoal_state.num_F; i++ ) {
        ft = ggoal_state.F[i];

        if ( lnum_clauses == MAX_CLAUSES ) {
            printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
            exit( 1 );
        }
        lclause[lnum_clauses] = ( int * ) calloc( 1, sizeof( int ) );
        if ( fcode[layer][ft] == -1 ) {
            printf("\n\ngoal fact not coded - level too low??\n\n");
            result = UNSAT;
            goto freeandquit;
        }
        lclause[lnum_clauses][0] = fcode[layer][ft];
        lclause_size[lnum_clauses] = 1;
        lnum_clauses++;
    } /* goals */

	
	int mutex_begin_pos  = lnum_clauses; 
	mutex_count = 0;
	long_in_mutex_count = 0;	
    /* 
     * exclusion constraints.
     */
    /* START CHANGE */
#if(DEBUG)
	printf("exclusion constraints layer [0 - %d]...\n",layer );
#endif
	for ( t = 0; t <= layer; t++ ) {

        for ( i1 = gin_fts_at[t]; i1 && i1->next; i1 = i1->next ) {
            i = i1->i1;
            for ( i2 = i1->next; i2; i2 = i2->next ) {
                j = i2->i1;
                if ( !(gft_conn[i].info_at[t]->bit_exclusives[gft_conn[j].uid_block] &
                        gft_conn[j].uid_mask) ) {
                    continue;
                }
             
				if ( f_f_mutex[i][j]!=-1 || f_f_mutex[j][i]!=-1 ){
					//long_in_mutex_count++;
					//mutex_count++;
					//continue;
				}

                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                if ( fcode[t][i] == -1 ) {

                    fcode[t][i] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    fop_to_code[code-1] = i;
                    ftime_to_code[code-1] = t;
                	code_type[code-1] = CODE_TYPE_FT;
                	code_cost[code-1] = 0;
				}
                lclause[lnum_clauses][0] = (-1) * fcode[t][i];
                if ( fcode[t][j] == -1 ) {

                    fcode[t][j] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    fop_to_code[code-1] = j;
                    ftime_to_code[code-1] = t;
                	code_type[code-1] = CODE_TYPE_FT;
                	code_cost[code-1] = 0;
				}
                lclause[lnum_clauses][1] = (-1) * fcode[t][j];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
                mutex_count++;
                
                //By R.Huang
                /* #1. if i is true at time t, then any action $a$ that 
                 * 		has j as precondtion cannot be true at time $t$;
                 * 
                 * #2. also if j is true, then similar case;
                 */
       			
       			//commented?????!!!
            } /* i2 */
        } /* i1 */

        if ( t == layer ) continue;
        /* END CHANGE */

        for ( i1 = gin_ops_at[t]; i1 && i1->next; i1 = i1->next ) {
            i = i1->i1;
            for ( i2 = i1->next; i2; i2 = i2->next ) {
                j = i2->i1;
                if ( !interfere( i, j ) ) {
                    continue;
                }
                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                if ( lcode[t][i] == -1 ) {

                    lcode[t][i] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    lop_to_code[code-1] = i;
                    ltime_to_code[code-1] = t;
                	code_type[code-1] = CODE_TYPE_OP;
                	//Q. Lv 2008-4-4
					code_cost[code-1] = gop_conn[i].cost;
				}
                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                if ( lcode[t][j] == -1 ) {

                    lcode[t][j] = code++;
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                    lop_to_code[code-1] = j;
                    ltime_to_code[code-1] = t;
                	code_type[code-1] = CODE_TYPE_OP;
                	//Q. Lv 2008-4-4
					code_cost[code-1] = gop_conn[j].cost;	
				}
                lclause[lnum_clauses][1] = (-1) * lcode[t][j];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            }
        }
    }


	/* *****************************************
	 * 
	 * londex constraints; Note, Only Facts
	 * 
	 * *****************************************/

	int t2,lit1;
	int londex_begin_pos = lnum_clauses;

	londex_pair_count = 0;
	londex_count = 0;
	
	for(i=0;i<gnum_ft_conn;i++)
	for(j=0;j<gnum_ft_conn;j++){
		if( i==j || f_f_mutex[i][j]==-1 ) continue;	
		londex_pair_count++;
		//Set constraint from tg->element[i] ---->   tg->element[j];
		for( t =0  ; t< layer; t++ ){
			if( fcode[t][i] == -1 ) continue;
			
			for(k=0;k<f_f_mutex[i][j];k++){
				t2 = t + k;
				if( t2 ==layer ) break;
				if( fcode[t2][j] == -1 ) continue;
				
				lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
				lclause_size[lnum_clauses] = 2;
				lclause[lnum_clauses][0] = -1 * fcode[t][i];
				lclause[lnum_clauses][1] = -1 * fcode[t2][j];
				lnum_clauses++;
				londex_count++;
				
				/*
				if(k!=0 && gcmd_line.londexm ){
					for(l=0;l<gft_conn[i].num_A;l++){
						t2 = t-k-1;					
						if( t2>=layer || t2 <0 ) break;
						if( lcode[t2][gft_conn[i].A[l]]==-1 ) continue;
						
	                	lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
										lclause[lnum_clauses][0] = (-1) * fcode[t][j];

						lclause[lnum_clauses][1] = (-1) * lcode[t2][gft_conn[i].A[l]];
		                lclause_size[lnum_clauses] = 2;
		                lnum_clauses++;
					}
				}
				*/
						
			}
		}
	}
	#if(DEBUG)	
	printf("\n");
	printf("Info:  Londex constraint(fact only) totally %d\n",londex_count);
	printf("       Londex constraint(fact pair) totally %d\n",londex_pair_count);
	printf("       Mutex constraints(fact only) totally %d\n",mutex_count);
	printf("       Overlap of two types(fact only) totally %d\n",long_in_mutex_count);
	/* End of Londex Encoding of fact; */
	#endif 
	 
    /* 
     * that's it. print CNF file.
     */
    if ( gcmd_line.debug ) {
			/* debugging: print semantic version of clause set.
			 */
	
	        if ( (VARFILE = fopen( gcmd_line.varFileName, "w" )) == NULL ) {
	            printf("\n\nEXIT: can't open VARFILE file.\n\n");
	            exit( 1 );
	        }
			fprintf(VARFILE, "\n\nc DECISION LAYER %d, THIN GP-BASED ENCODING.", layer);
			fprintf(VARFILE, "\n\nc VARS:");
	
	        for ( i = 1; i < code; i++ ) {
	            fprintf(VARFILE, "\nc var %2d action (", i);
	  	    print_op_nameToFile( lop_to_code[i], VARFILE );
		    fprintf(VARFILE, ") %d", ltime_to_code[i] );
			}

		/*
		printf("\n\nCLAUSES:");
		for ( i = 0; i < lnum_clauses; i++ ) {
		    printf("\n%6d - ", i);
		    if ( 1 ) {
			for ( j = 0; j < lclause_size[i]; j++ ) {
			    printf("(%d)", lclause[i][j]);
			    if ( j < lclause_size[i]-1 ) printf(", ");
			}
			fflush(stdout);
		    }
		    for ( j = 0; j < lclause_size[i]; j++ ) {
			if ( lclause[i][j] > 0 ) {
			    printf("(");
			    print_op_name( lop_to_code[lclause[i][j]] );
			    printf(" at %d)", ltime_to_code[lclause[i][j]] );
			} else {
			    printf("(NOT ");
			    print_op_name( lop_to_code[(-1)*lclause[i][j]] );
			    printf(" at %d)", ltime_to_code[(-1)*lclause[i][j]] );
			}
			if ( j < lclause_size[i]-1 ) printf(", ");
		    }
		}
		*/
        fclose(VARFILE);
    }

#if(DEBUG)	
    printf("\n\nDECISION LAYER %d, WRITING THIN GP-BASED ENCODING WITH %d VARS, %d CLAUSES.\n",
	   layer, code-1, lnum_clauses);
	//This commented is to write encoding to CNF file;
    if ( create == 1 ) {
        if ( (CNF = fopen(gcmd_line.cnfFileName, "w")) == NULL ) {
	    	printf("\n\ncan't open CNF file.\n\n");
            exit( 1 );
        }

        fprintf(CNF, "c CNF file planning task -p %s, -o %s, -f %s, bound %d, thin gp-based encoding\n",
	        gcmd_line.path, gcmd_line.ops_file_name, gcmd_line.fct_file_name, layer);

        fprintf(CNF, "p cnf %d %d\n", code-1, lnum_clauses);
        for ( i = 0; i < lnum_clauses; i++ ) {
		    for ( j = 0; j < lclause_size[i]; j++ ) {
		        fprintf(CNF, "%d ", lclause[i][j]);
		    }
		    fprintf(CNF, "0\n");
        }
        fclose(CNF);
    }
#endif
    
    //This is for writing encoding to MINISAT solver structure;
    if ( create == 0 ) {
#if(DEBUG)
		printf("c CNF file planning task -p %s, -o %s, -f %s, bound %d, thin gp-based encoding\n",
			gcmd_line.path, gcmd_line.ops_file_name, gcmd_line.fct_file_name, layer);
      
		printf("In building up solver: p cnf %d %d\n", code-1, lnum_clauses);
		fflush(stdout);
#endif
		veci tmp_lits;
    	veci_new(&tmp_lits);
    	
		for ( i = 0; i <lnum_clauses ; i++ ) {
	        lit* begin;
			veci_resize(&tmp_lits,0);
		  	
		  	for ( j = 0; j < lclause_size[i]; j++ ) {
		  		int var;
				var = abs(lclause[i][j])-1;
				veci_push(&tmp_lits, (lclause[i][j] > 0 ? toLit(var) : lit_neg(toLit(var))));
			}
			
			begin = veci_begin(&tmp_lits);
	    	if (!solver_addclause( minisat, begin, begin+veci_size(&tmp_lits))){
	        	//veci_delete(&tmp_lits);
	        	//printf("CRITICAL ERROR! CANNOT ADD CLAUSE ANY MORE.\n");
	        	//exit(-1);
	        }
			free(lclause[i]);
			lclause[i]=NULL;
  		}
		free(lclause);
		lclause = NULL;
		veci_delete(&tmp_lits);
		//Q. Lv 2008-4-3	
		for( i = 0; i < code-1 ;i++){
			minisat->var_type[i] = code_type[i+1];
			minisat->var_cost[i] = code_cost[i+1];
		}
		set_priority(minisat, layer);
	}
    
freeandquit:
	minisat->lop_to_code = lop_to_code;
	minisat->ltime_to_code = ltime_to_code;
	lop_to_code = NULL;
	ltime_to_code = NULL;
    post_solving(layer);fflush(stdout);
    return result;
}

void set_priority(solver* s, int layer)
{
	printf("\nset priority... ");fflush(stdout);
	if(s->priority==NULL)
		return;
	assert(s->priority!=NULL);
	
	double* activity = s->activity;
	double* priority = s->priority;
	int ft, var;
	for(int i = 1 ; i < layer ; i++){
		//printf("\nlayer: %d",i);
		for(int j = 0; j < ggoal_state.num_F; j++){
			ft = ggoal_state.F[j];
			if(fcode[i][ft]!=-1){
				var = fcode[i][ft] - 1 ;
				priority[var] = i + gmax_action_cost + 1; 
		//		printf("var %d--priority %lf\n",var,priority[var]);fflush(stdout);
			}
		}
	}
	for(int i = 0; i < layer; i++){
		for( int j = 0; j < gnum_op_conn; j++){
			if(lcode[i][j]!=-1){
				var = lcode[i][j] - 1;
				priority[var] = gop_conn[j].cost;
		//		printf("var %d--priority %lf\n",var,priority[var]);fflush(stdout);
			}
		}
	}
	
#if(DEBUG)
	printf("done!\n");fflush(stdout);
#endif
	return ;
}

void post_solving(int curr_layer)
{
	int i;
	if(lclause_size!=NULL) free(lclause_size);
  	if(lclause){
  		for(i=0;i<lnum_clauses;i++)
  			if(lclause[i]) free(lclause[i]);
  		free(lclause);
  	}
  	
  	/* action coding */
  	if(lcode!=NULL){
  		//for(i=0;i<gcmd_line.cnflayer;i++)
  		for(i=0;i<curr_layer;i++)
			if (lcode[i]!=NULL) free(lcode[i]);
  		free(lcode);
  	}
  	if(lop_to_code!=NULL) free(lop_to_code);
  	if(ltime_to_code!=NULL) free(ltime_to_code);
  	
  	/* fact coding*/
  	if(fcode!=NULL){
  		for(i=0;i<curr_layer;i++)
			if (fcode[i]!=NULL) free(fcode[i]);
  		free(fcode);
  	}
  	if(fop_to_code!=NULL) free(fop_to_code);
  	if(ftime_to_code!=NULL) free(ftime_to_code);
  	  	  	
  	if(F!=NULL) free(F);
  	if(A!=NULL) free(A);
  	
  	free(code_type);
	free(code_cost);
}

void printStats(stats* stats, int cpu_time)
{
    double Time    = (float)(cpu_time)/(float)(CLOCKS_PER_SEC);
    printf("\nrestarts          : %12d\n", stats->starts);
    printf("conflicts         : %12.0f           (%9.0f / sec      )\n",  (double)stats->conflicts   , (double)stats->conflicts   /Time);
    printf("decisions         : %12.0f           (%9.0f / sec      )\n",  (double)stats->decisions   , (double)stats->decisions   /Time);
    printf("propagations      : %12.0f           (%9.0f / sec      )\n",  (double)stats->propagations, (double)stats->propagations/Time);
    printf("inspects          : %12.0f           (%9.0f / sec      )\n",  (double)stats->inspects    , (double)stats->inspects    /Time);
    printf("conflict literals : %12.0f           (%9.2f %% deleted  )\n", (double)stats->tot_literals, (double)(stats->max_literals - stats->tot_literals) * 100.0 / (double)stats->max_literals);
	//Q. Lv 2008-4-9
	printf(stats->solution_num > 0? "SATISFIABLE\n" : "UNSATISFIABLE\n" );
	if(stats->solution_num > 0){
		printf("have found %d solutions\n",stats->solution_num);
		printf("current layer %d ",gcurr_layer);
		printf("current minimal costs :%d\n",gmini_costs);
	}
	printf("current time :%7.2lf\n",get_global_time());
	fflush(stdout);
}



/*
 * Return value is sat or not;
 */
void *execute_solver()
{
	int create = 0;
	
	solver *minisat = solver_new();	
	int curr_layer = gcmd_line.cnflayer;
	int is_code = do_cnf_output(minisat,create);	
	
	if ( gcmd_line.londex_mode!=LONDEX_MODE_NONE ) 		
	memory_usage_evaluation(minisat);
	if (gcmd_line.londex_mode == LONDEX_MODE_STATIC ) 
		generate_static_londex(minisat);

	if ( is_code!=SAT ){
		solver_delete(minisat);
		return NULL;
	}
	int     clk = clock();

	check_memory_usage();
    
#if(DEBUG)
	print_sat_solver_to_file(minisat);
#endif
	minisat->verbosity = 1;
	solver_solve( minisat , 0 , 0 );
	
	check_memory_usage();
	printStats(&minisat->stats, clock() - clk);
	solver_delete(minisat);
	return NULL;
}


void *thread_solve()
{
	int create = 0;
#if(!DEBUG)
	pthread_mutex_lock(&layer_lock);

	gcmd_line.cnflayer++;
	solver *minisat = solver_new();	
	int curr_layer = gcmd_line.cnflayer;
	printf("\n\n********  Begin solve with multi-thread at %d layer  ********\n\n",curr_layer);fflush(stdout);
	int is_code = do_cnf_output(minisat,create);	
	
	if ( gcmd_line.londex_mode!=LONDEX_MODE_NONE ) 		
		memory_usage_evaluation(minisat);
	if (gcmd_line.londex_mode == LONDEX_MODE_STATIC ) 
		generate_static_londex(minisat);

	
	pthread_mutex_unlock(&layer_lock);
	
	if ( is_code!=SAT ){
		solver_delete(minisat);
		return NULL;
	}
	int     clk = clock();

    
	minisat->verbosity = 0;
	solver_solve( minisat , 0 , 0 );
	
	printStats(&minisat->stats, clock() - clk);
	printf("\n*******  End solve with multi-thread at %d layer  ********\n\n",curr_layer);fflush(stdout);
	solver_delete(minisat);
	pthread_exit(NULL);
#endif
}


/**************************************************
 * 
 * 			Londex Constraints for SAT solving;
 * 
 * ************************************************/

void memory_usage_evaluation(solver *s){
#ifndef OS_CYGWIN
	float MAX_MEMORY_BEFORE_SOVING = 1000.00;
	float consumption = getVmRSS();	
	consumption  = MAX_MEMORY_BEFORE_SOVING - consumption;
	if( consumption<0 ) consumption = 0;
	s->clauseAddingLimit = (int)(consumption * 1024 * 1024 /32) ;
#endif
}



/*
 * if no conflict then return -1;
 * 		otherwise, return lit_var
 */
/*
clause* apply_constraints(solver* s,  lit l){ 
	
	int     clk = clock();
	int i,j;
	int var = lit_var(l);
	if( code_type[var+1] != 0 ) return (clause*)0; //0 means fact, 1 means action
	
	struct DisAction* index = NULL;
	clause* confl = (clause*)0;
	veci tmp_clause;

	int fact = fop_to_code[var+1]; 		//Get Action for variable i
	int level = ftime_to_code[var+1];   //Get Time_level for variable i	
	
	int *infer_list = malloc( sizeof(int) * s->size * 2 );
	int infer_list_size = 0;

	for(i=0;i<gnum_ft_conn;i++){
		for(  j=0   ; j<f_f_mutex[fact][i]; j++ ){
			//if( (level+j) >= gcmd_line.cnflayer) break;
			if( (level+j) >= s->stats.layer) break;
			if( fcode[level+j][i] == -1 ) continue;
			
			int restricted_var = fcode[level+j][i] -1 ;
			assert( restricted_var>=0 && restricted_var<s->size );

			//set restrited_var as invalid in assignments;
			if ( s->assigns[restricted_var] == l_Undef )
				infer_list[infer_list_size++] = restricted_var;
			else if ( s->assigns[restricted_var] == l_True ){

				veci_new(&tmp_clause);
				veci_push( &tmp_clause, lit_neg(toLit(restricted_var)));
				veci_push( &tmp_clause, lit_neg(l) );

				clause* c = clause_new( s, veci_begin(&tmp_clause),veci_begin(&tmp_clause)+veci_size(&tmp_clause), 0 );
				veci_delete(&tmp_clause);					
				apply_generate++;
				vecp_push( &s->clauses , c );

				if(confl==0){
					confl = s->binary;
					(clause_begin(confl))[1] = lit_neg(l);
					(clause_begin(confl))[0] = lit_neg(toLit(restricted_var));
					goto confl_found;						
				}
			}
		}
	}


confl_found:
	if( confl!=0 ){
		free(infer_list);
		return confl;
	}

	for(j=0;j<infer_list_size;j++)
		enqueue(s,lit_neg(toLit(infer_list[j])), clause_from_lit(l) );

	apply_count++;
	free(infer_list);
	clk = clock() - clk;
	londex_check_time_long += (float)clk / (float)CLOCKS_PER_SEC;
	return confl;	//everything is OK;
}
*/
/*
 *
 * 
 * 	Action-Based Static 
 * 
 */
void generate_static_londex_ab_encoding(solver *s){
	
	printf("Generating static londex[action-based]...\n");
	int i,t,action,j,k;
	struct DisAction* index = NULL;
	lit* begin;

	veci tmp_lits;
	veci_new(&tmp_lits);

    int count = 0;
    for(j=0;j<20;j++){
		for(action=0;action<gnum_op_conn;action++){
			if( A[action] == -1 ) continue;
			
			//for( t = gcmd_line.cnflayer-1 ; t>=0; t-- ){	
			for( t = s->stats.layer-1 ; t>=0; t-- ){	

				if( A[action]>t || lcode[t][action] == -1) continue;
				
				int lit1 = lcode[t][action] - 1;
				for(index = distance_action[action].next;index;index=index->next){
					
					if(j>=index->value) continue;
						
					int t2 = t+j;
					//if(t2>=gcmd_line.cnflayer) break;  //Should break, otherwise redundant clause;
					if(t2 >= s->stats.layer) break; 
					if ( A[index->actionID] == -1 || A[index->actionID] > t2 ) continue; 
					if( lcode[t2][index->actionID] == -1 ) continue;					
					
					int restricted_var = lcode[t2][index->actionID] -1 ;
					veci_resize( &tmp_lits,0);
					veci_push(&tmp_lits, lit_neg(toLit(lit1)));			//Only lit_neg here can make it right??? Why ????
					veci_push(&tmp_lits, lit_neg(toLit(restricted_var)));
					begin = veci_begin( &tmp_lits );
					
					solver_addclause( s, begin, begin + veci_size(&tmp_lits));
					if (count++ > s->clauseAddingLimit) 
						goto DONE;
				}
			}
		}
    }
	
DONE:
	veci_delete(&tmp_lits);
	printf("Totally %d clauses added in static method.\n",count);
#ifndef OS_CYGWIN
	printf("memory cost is %lf\n", getVmRSS());
#endif

	//This function is for debugging;
	//if(gcmd_line.printEncoding!=-1 && gcmd_line.printEncoding == gcmd_line.cnflayer){
	if(gcmd_line.printEncoding!=-1 && gcmd_line.printEncoding == s->stats.layer){
		//printf("Print cnf encoding %d\n %d ,%d\n", gcmd_line.cnflayer, gcmd_line.printEncoding,  gcmd_line.cnflayer);
		printf("Print cnf encoding %d\n %d ,%d\n", s->stats.layer, gcmd_line.printEncoding,  s->stats.layer);
		print_sat_solver_to_file(s);
		exit(0);
	}
	//printf("Totally %d clauses added in static londex(non-binary).\n",count);
}


/*
 * 	Graph-Based
 * 
 * This methods is to add constraints for gp_based encoding(cnfout type 4), 
 * 	Two kinds of constraints are added, both for *actions*; 
     *facts* is added in cnfout.c
 */
void generate_static_londex_gp_encoding(solver *s){
	
	printf("Generating static londex[graph-based]...\n");fflush(stdout);
	int i,t,action,j,k,t2,lit1,restricted_var;
	struct DisAction* index = NULL;
	lit* begin;

	veci tmp_lits;
	veci_new(&tmp_lits);
    int count = 0;
	
    /************* Action constraints for Graph-based encoding ***********/
	for(action=0;action<gnum_op_conn;action++){
		
		//for( t = gcmd_line.cnflayer-1 ; t>=0; t-- ){
		for( t = s->stats.layer-1 ; t>=0; t-- ){
			if( lcode[t][action] == -1) continue;
			
			lit1 = lcode[t][action] - 1;
			for(index = distance_action[action].next;index;index=index->next){
				
				for(j=0;j<index->value;j++){

					t2 = t+j;
					//if( t2>=gcmd_line.cnflayer ) break;  //Should break, otherwise redundant clause;
					if( t2 >= s->stats.layer ) break; 
					if( lcode[t2][index->actionID] == -1 ) continue;
					
					int restricted_var = lcode[t2][index->actionID] - 1;
					veci_resize( &tmp_lits, 0 );
					veci_push( &tmp_lits, lit_neg(toLit(lit1)) );
					veci_push( &tmp_lits, lit_neg(toLit(restricted_var)) );
					begin = veci_begin( &tmp_lits );
					
					solver_addclause( s, begin, begin + veci_size(&tmp_lits));
					count++;
#ifndef OS_CYGWIN
					if ( count > s->clauseAddingLimit ) 
						goto DONE;
#endif

				}
			}

		}
	}
	
DONE:
	printf("Totally %d clauses added in static method.\n",count);fflush(stdout);
	veci_delete(&tmp_lits);
#ifndef OS_CYGWIN
	printf("memory cost is %lf\n", getVmRSS());
#endif

	//This function is for debugging;
	//if(gcmd_line.printEncoding!=-1 && gcmd_line.printEncoding == gcmd_line.cnflayer){
	//	printf("Print cnf encoding %d\n %d ,%d\n", gcmd_line.cnflayer, gcmd_line.printEncoding,  gcmd_line.cnflayer);
	if(gcmd_line.printEncoding!=-1 && gcmd_line.printEncoding == s->stats.layer){
		printf("Print cnf encoding %d\n %d ,%d\n", s->stats.layer, gcmd_line.printEncoding,s->stats.layer);
		print_sat_solver_to_file(s);
		exit(0);
	}
}


void generate_static_londex(solver *s){
	//printf("Generating Static Londex(Action):");fflush(stdout);
	if(gcmd_line.cnfout==4)
		generate_static_londex_gp_encoding(s);
	else if (gcmd_line.cnfout==1)
		generate_static_londex_ab_encoding(s);
	else
		printf("No static londex method implemented for this encoding.\n"),exit(1); 
}



void validate_londex_ab_encoding(solver* s){
	int var,i;
	struct DisAction* index = NULL;
	
	for(var=0;var<s->model.size;var++){
		if( s->model.ptr[var] == -1 ) continue;
			else if ( s->model.ptr[var] == 0 ) {printf("Strange, quit\n");exit(0);}
		//printf(">>>>>Checking variable %d\n",var+1);

		int action = lop_to_code[var+1]; //Get Action for variable i
		int level = ltime_to_code[var+1];   //Get Time_level for variable i
		//printf("     Action "),print_op_name(action),printf("[level%d]\n",level);

		for(index = distance_action[action].next;index;index=index->next){
			//printf("                         %d",index->value),print_op_name(index->actionID),printf("\n");
			for(i=0;i<index->value;i++){
				//if( (level+i) >= gcmd_line.cnflayer)  break;
				if( (level+i) >= s->stats.layer)  break;
				if ( A[index->actionID] == -1 || A[index->actionID] > level+i ) continue; 
				if( lcode[level+i][index->actionID] == -1 ) continue;
									
				int restricted_var = lcode[level+i][index->actionID] -1 ;
				
				assert( restricted_var>= 0 && restricted_var < s->size );
				
				//set restrited_var as invalid in minisat;
				if ( s->model.ptr[restricted_var] == 1 ){
					printf("\n WRONG:");
					print_op_name( action );
					printf("[%d][%d]\t\t", action, level );
					print_op_name( index->actionID ),printf( "[%d]cannot at:[%d]",action, level+i );
				}
			}
		}
	}
	
	printf("validation for action-based encoding finished.\n");
}

/* Action
void validate_londex_gb_encoding(solver* s){

	int var,i,j,k,value;
	struct DisAction* index = NULL;
	printf("Start of validation of gb encoding solving result;\n");
	
	for(var=0;var<s->model.size;var++){
		if( s->model.ptr[var] == -1 ) continue;
		else if ( s->model.ptr[var] == 0 ){
			printf("Strange, quit\n");exit(0);
		}
		if( code_type[var+1] == 1 ) continue;
		//printf("\n>>>>>Checking variable %d\n",var+1);

		int action = lop_to_code[var+1]; //Get Action for variable i
		int level =  ltime_to_code[var+1];   //Get Time_level for variable i
		//printf("Action:");print_op_name(action);printf(";level:%d\n",level);
		
		for(index = distance_action[action].next;index;index=index->next){
			for(i=0;i<index->value;i++){
				if( (level+i) >= gcmd_line.cnflayer )  break;

				if( lcode[level+i][index->actionID] == -1 ) continue;
				int restricted_var = lcode[ level+i ][index->actionID] -1 ;
				assert( restricted_var >= 0 && restricted_var < s->size );
				
				//set restrited_var as invalid in minisat;
				if ( s->model.ptr[restricted_var] == 1 ){
					printf("WRONG:");
					print_op_name( action );
					printf("[%d][%d]\n", action, level );
					print_op_name( index->actionID );
					printf( "[%d]cannot at:[%d]\n", index->actionID, level+i );
				}
			}
		}
		
	}
		
	printf("Validation for graph-based encoding solution finished.\n");
}
*/

void validate_londex_gb_encoding(solver* s){

	int var,i,j,k,value,g;
	struct DisAction* index = NULL;
	printf("Start of validation of gb encoding solving result;\n");
	
	for(i=0;i<gnum_ft_conn;i++)
	for(j=0;j<gnum_ft_conn;j++)
		f_f_mutex[i][j] = -1;
		
	for(g=0;g<transGraphNum;g++){
		
		TransitionGraph* tg = &(transGraph[g]);
		for(i=0;i<tg->elementSize;i++)
		for(j=0;j<tg->elementSize;j++){
			if(i==j) continue;
			
			int eI = tg->element[i];
			int eJ = tg->element[j];
			
			int value =  tg->londexValue[i][j] ;

			if( f_f_mutex[eI][eJ] == -1)
				f_f_mutex[eI][eJ] = value;
			else if ( f_f_mutex[eI][eJ] > value )
				f_f_mutex[eI][eJ] = value;
		}
	}
		
	for(var=0;var<s->model.size;var++){
		if( s->model.ptr[var] != 1 ) continue;

		if( code_type[var+1] == 1 ) continue;
		//printf("\n>>>>>Checking variable %d\n",var+1);

		//int action = lop_to_code[var+1]; //Get Action for variable i
		int fact = fop_to_code[var+1];
		int level =  ftime_to_code[var+1];   //Get Time_level for variable i
		
		for(i=0;i<gnum_ft_conn;i++){
			if( f_f_mutex[fact][i]<1 ) continue;
			
			for(j=0;j<f_f_mutex[fact][i];j++){
				//if( (level+j) == gcmd_line.cnflayer )  break;
				if( (level+j) == s->stats.layer )  break;

				int restricted_var = fcode[level+j][i]-1;
				//if(   restricted_var < 0 || restricted_var >= s->model.size  )  continue;
				
				//printf("verifying variable %d and %d", var, restricted_var );
				if( s->model.ptr[restricted_var] == 1){
					printf("This constraint is WRONG:");
					print_ft_name( fact );
					//printf("[%d][%d]\n", fact, );
					printf(" &&&& ");
					print_ft_name( i );
					printf( " distance at least %d\n", f_f_mutex[fact][i] );
					printf(" however, they are found to have distance of %d\n",j);
				}				
			}
		}
		/*
		for(i=0;i<gnum_ft_conn;i++){
			if( f_f_mutex[i][fact]<1 ) continue;
			
			for(j=0;j<f_f_mutex[i][fact];j++){
				if( (level-j) < 0 )  break;
				
				int restricted_var = fcode[level-j][i]-1;
				//if(   restricted_var < 0 || restricted_var >= s->model.size  )  continue;
				
				//printf("verifying variable %d and %d", var, restricted_var );
				if( s->model.ptr[restricted_var] == 1){
					printf("This constraint is WRONG:");
					print_ft_name( fact );
					//printf("[%d][%d]\n", fact, );
					printf(" &&&& ");
					print_ft_name( i );
					printf( " distance at least %d\n", f_f_mutex[i][fact] );
					printf(" however, they are found to have distance of %d\n",j);
				}				
			}
		} */		
	}
		
	printf("Validation for graph-based encoding solution finished.\n");
}


void validate_model_with_londex(solver* s){

	if(gcmd_line.cnfout==4)
		validate_londex_gb_encoding(s);
	else if (gcmd_line.cnfout==1)
		validate_londex_ab_encoding(s);
	else
		printf("No static londex method implemented for this encoding.\n"),exit(1); 
}
