// pretreat.cpp : Defines the entry point for the console application.
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "pretreat.h"
#include "bb.h"

char line[MAX_LINE_LEN] =""; 
char* to_read;
char yyltext[MAX_LINE_LEN]="";
int yylval;

void read_line(char *line, FILE *ifp) 
{
	fgets(line,MAX_LINE_LEN,ifp);
	if (line[MAX_LINE_LEN-1]=='\0') {
		printf("exceeds max length of %d!\n",MAX_LINE_LEN-2);
		exit(1);
	}
}

int key_word(char **in)
{
	char tmp[MAX_LENGTH] = "";
	int i=0;
	
	while(**in==' ' || **in=='\t'){
		(*in)++;
	}

	while(**in!=' ' && **in!='\t' && **in !='\n' && **in!=')' && **in != '.'){
		yyltext[i++] = **in;
		(*in)++;
	}
	tmp[i] = '\0';
	yyltext[i] = '\0';

	if(strcmp(yyltext,":requirements")==0)
		return REQUIREMENTS_TOK;
	if(strcmp(yyltext,":action-costs")==0)
		return ACTION_COSTS_TOK;
	else if(strcmp(yyltext,":action")==0)
		return ACTION_TOK;
	else if(strcmp(yyltext,":metric")==0)
		return METRIC_TOK;
	else if(strcmp(yyltext,":functions")==0)
		return FUNCTIONS_TOK;
	else if(strcmp(yyltext,"increase")==0)
		return INCREASE_TOK;
	else if(strcmp(yyltext,"total-cost")==0)
		return TOTAL_COST_TOK;
	else if(strcmp(yyltext,"minimize")==0)
		return MINIMIZE_TOK;
	else if(strcmp(yyltext,"number")==0)
		return NUMBER_TOK;
	else if(yyltext[0]==':'&&yyltext[1]=='f'&&yyltext[2]=='u')
		return FUNCTIONS_TOK;

	return OTHERS_TOK;
}

void skip_space(char** in)
{
	while(**in==' ' || **in=='\t'){
		(*in)++;
	}
}

void skip_backToken(char** in)
{
	while(**in==' ' || **in=='\t'){
		(*in)--;
	}
	while(**in!=' '&& **in!= '\t')
		(*in)--;

}

int get_Token(char** in)
{

	while(**in==' ' || **in=='\t' || **in=='\r'){
		(*in)++;
	}
	
	if(**in=='('){
		(*in)++;
		return OPEN_PAREN;
	}
	if(**in==')'){
		(*in)++;
		return CLOSE_PAREN;
	}
	if(**in==';'){
		(*in)++;
		return COMMENT_TOK;
	}
	if(**in=='\n'){
		return LINEEND_TOK;
	}

	return key_word(in);

}

int get_backParen(char** in)
{
	while(**in!='(')
		(*in)--;
	if(**in=='('){
		(*in)++;
		return OPEN_PAREN;
	}
	return OTHERS_TOK;

}
int get_backEqual(char** in)
{
	while(**in!='=')
		(*in)--;
	if(**in=='='){
		(*in)++;
		return EQUAL_TOK;
	}
	return OTHERS_TOK;	
	
}

void pretreat_ops_file(char* filein, char* fileout)
{
	//my_m is our main data structure that will be used throughout the execution
	unsigned long i= 0;
  
  //line is a buffer for reading input
	char line1[MAX_LINE_LEN];
	line1[MAX_LINE_LEN-1] = '\n'; 
	char tmp[MAX_LINE_LEN];
	tmp[MAX_LINE_LEN-1] = '\n';

	//char** in;
	char c;
	int token,j;
	int length;
	FILE *fin, *fout;

	line[MAX_LINE_LEN-1] = '\n';
	
	if ( (fin = fopen( filein, "r" )) == NULL ) {
        	printf("\n\nEXIT: can't open % file.\n\n",filein);
 		exit(1);
    	}

	if ( (fout = fopen( fileout, "w" )) == NULL ) {
        printf("\n\nEXIT: can't open pretreatment %s file.\n\n",fileout);
 		exit(1);
	}
#if(DEBUG)
	printf("begin to pretreat domain file... ");fflush(stdout);
#endif
  	//main parsing loop
	while((c=getc(fin))!=EOF){
		//if(isspace(c)) continue; //skip any space at the beginning of the line
		ungetc(c,fin);

		read_line(line,fin);
		to_read = line;
		//in = &to_read;

		while( (*to_read) != '\n' ){
			token = get_Token(&to_read);
			if(token==COMMENT_TOK)
				break;
			//delete action-costs
			if(token==ACTION_COSTS_TOK){
				gaction_costs_def = TRUE;
#if(DEBUG)
				printf("action-costs defined.\n");fflush(stdout);
#endif
				length = strlen(":action_costs");
				//c = **in;
				for(j = length;j >0; j--){
					*(to_read-j) = ' ';
				}
				//break;
			}
			//delete :functions total-cost
			if(token==FUNCTIONS_TOK){
				//printf("functions list is empty!\n");
				if((token=get_Token(&to_read))==LINEEND_TOK ){
					for(i=0;i<MAX_LINE_LEN;i++)
						tmp[i]=line[i];
				if(get_nextToken(&to_read,fin)==OPEN_PAREN && get_nextToken(&to_read,fin)==TOTAL_COST_TOK){
					get_nextToken(&to_read,fin);
					get_nextToken(&to_read,fin);
					get_nextToken(&to_read,fin);//skip ") - number "
					if(get_nextToken(&to_read,fin)!=CLOSE_PAREN)
						fputs(tmp,fout);
					else{

						get_nextToken(&to_read,fin);
						get_nextToken(&to_read,fin);
					}
				}else
					fputs(tmp,fout);
				}
				
			}

			if(token==TOTAL_COST_TOK){
				//printf("delete total-cost!\n");
				
				if(get_backParen(&to_read)==OPEN_PAREN){
					char* bound = to_read - 1;
					get_Token(&to_read);//skip total-cost
					get_Token(&to_read);// skip )
					token = get_Token(&to_read);
					while(token != CLOSE_PAREN && token != LINEEND_TOK) 
						token = get_Token(&to_read);
					
					if(token==CLOSE_PAREN){
						bound = line;
					}

					while(bound!=to_read){
						*bound = ' ';
						bound++;
					}
					/*
					if(token == LINEEND_TOK){
						*bound = ' ';
						
						read_line(line1,fin);
						bound = line + strlen("(:functions") +1;
						j = 0;
						while(line1[j]!= '\n'){
							*(bound++) = line1[j++];	
						}
						*bound = '\n';
					}
					*/
					break;

				}
			}
			//delete increase total-cost
			if(token==INCREASE_TOK){
				if(get_backParen(&to_read)==OPEN_PAREN){
					char* bound = to_read - 1;
					int bracket =1;
					while(1){
						token = get_Token(&to_read);
						while(token != OPEN_PAREN && token !=CLOSE_PAREN) 
							token = get_Token(&to_read);
						if(token==OPEN_PAREN)
							bracket++;
						else if(token==CLOSE_PAREN)
							bracket--;
						if(bracket==0){
							while((bound)!= to_read){
								*bound = ' ';
								bound++;
							}
							break;
						}
					}
				}
				//break;
			}

		}
		//puts(line);
		fputs(line,fout);
	}

	fclose(fin);
	fclose(fout);
#if(DEBUG)
	printf("done\n");
#endif

}

void pretreat_fct_file(char* filein, char* fileout)
{
	  //my_m is our main data structure that will be used throughout the execution
	unsigned long i= 0;
  
  //line is a buffer for reading input
	line[MAX_LINE_LEN-1] = '\n';
	//char** in;
	char c;
	int token,j;
	char* to_read;
	int length;
	FILE *fin, *fout;

	if ( (fin = fopen( filein, "r" )) == NULL ) {
        	printf("\n\nEXIT: can't open %s file.\n\n",filein);
 		exit(1);
    	}

	if ( (fout = fopen( fileout, "w" )) == NULL ) {
        printf("\n\nEXIT: can't open pretreatment %s file.\n\n",fileout);
 		exit(1);
	}
#if(DEBUG)
	printf("begin pretreat problem file... ");
#endif
  //main parsing loop
	while((c=getc(fin))!=EOF){
		//if(isspace(c)) continue; //skip any space at the beginning of the line
		ungetc(c,fin);

		read_line(line,fin);
		to_read = line;
		//in = &to_read;

		while( (*to_read) != '\n' ){
			token = get_Token(&to_read);
			if(token==COMMENT_TOK)
				break;

			//delete total-cost
			if(token==TOTAL_COST_TOK){
				//printf("init total cost!\n");fflush(stdout);
				if(get_backEqual(&to_read)==EQUAL_TOK){
					if(get_backParen(&to_read)==OPEN_PAREN){
						char* bound = to_read - 1;
						int bracket =1;
						while(1){
							token = get_Token(&to_read);
							while(token != OPEN_PAREN && token !=CLOSE_PAREN && token!= LINEEND_TOK) 
								token = get_Token(&to_read);
							if(token==OPEN_PAREN)
								bracket++;
							else if(token==CLOSE_PAREN)
								bracket--;
							if(bracket==0){
								while((bound)!= to_read){
									*bound = ' ';
									bound++;
								}
								break;
							}
						}
					}
				}else{
					to_read = line;
					while( get_Token(&to_read) != TOTAL_COST_TOK) ;

					if(get_backParen(&to_read)==OPEN_PAREN){
						char* bound = to_read -1;
						get_Token(&to_read);
						if(get_Token(&to_read)==CLOSE_PAREN){
							while((bound)!= to_read){
									*bound = ' ';
									bound++;
							}
						}
					}
				}
			}
			//delete metric total-cost
			if(token==METRIC_TOK){
				if(get_backParen(&to_read)==OPEN_PAREN){
					char* bound = to_read - 1;
					int bracket =1;
					while(1){
						token = get_Token(&to_read);
						if(token==OPEN_PAREN)
							bracket++;
						else if(token==CLOSE_PAREN)
							bracket--;

						if(token == TOTAL_COST_TOK){
							gmini_costs_require = TRUE;
#if(DEBUG)
							printf("metric minimize total-cost \n");fflush(stdout);
#endif
						}

						if(token == LINEEND_TOK){
							char* tmp = to_read;
						
							get_backParen(&to_read);
							to_read -= 2;
							skip_backToken(&to_read);							
							bound = to_read;
							to_read = tmp;
							break;
						}
						if(bracket==0){							
							break;
						}
					}
					while((bound)!= to_read){
								*bound = ' ';
								bound++;
							}
				}
				//break;
			}

		}
		//puts(line);
		fputs(line,fout);	
	}
	fclose(fin);
	fclose(fout);
}

int get_nextToken(char** in,FILE* fin)
{

	while(**in==' ' || **in=='\t'){
		(*in)++;
	}
	
	if(**in=='('){
		(*in)++;
		return OPEN_PAREN;
	}
	if(**in==')'){
		(*in)++;
		return CLOSE_PAREN;
	}
	if(**in==';'){
		(*in)++;
		return COMMENT_TOK;
	}
	if(**in=='\n'){
		read_line(line,fin);			
		to_read = line;
		return get_nextToken(&to_read,fin);
	}

	return key_word(in);

}

void scan_ops_file(char *ops_file)
{
	//my_m is our main data structure that will be used throughout the execution
	unsigned long i= 0;
  
  //line is a buffer for reading input
	line[MAX_LINE_LEN-1] = '\n';
	//char** in;
	char c;
	int token,j;
	int length;
	FILE *fin;

	if ( (fin = fopen( ops_file, "r" )) == NULL ) {
        	printf("\n\nEXIT: can't open % file.\n\n",ops_file);
 		exit(1);
    }
#if(DEBUG)
	printf("begin scan domain file... \n");
#endif
  	//main parsing loop
	while((c=getc(fin))!=EOF){
		//if(isspace(c)) continue; //skip any space at the beginning of the line
		ungetc(c,fin);

		read_line(line,fin);
		to_read = line;
		//in = &to_read;

		while( (*to_read) != '\n' ){
			token = get_Token(&to_read);
			if(token==COMMENT_TOK)
				break;
			if(token==ACTION_TOK){
				//get name of action
				token = get_nextToken(&to_read,fin);
				char action_name[MAX_LENGTH] ="";
				for(j = 0;j < strlen(yyltext); j++){
					if( yyltext[j]>='a' && yyltext[j]<='z' )
						action_name[j] = (char) ((int)yyltext[j]-32);
					else if(yyltext[j]=='-'||(isdigit(yyltext[j])))
						action_name[j] = yyltext[j];
				}
				action_name[j]='\0';
				int index = position_in_operators_table(action_name);
				if(index == -1){
					printf("can't find action name in operators table\n");fflush(stdout);
					printf("%s",action_name);fflush(stdout);
					break;//fclose(fin);
					//exit(1);
				}
				int bracket = 1;
					
				while( token != INCREASE_TOK && bracket > 0 ){
					token = get_nextToken(&to_read,fin);
					if(token==OPEN_PAREN)
						bracket++;
					if(token==CLOSE_PAREN)
						bracket--;
				}
				if(bracket==0){
					goperators[index]->cost = 0;
				}else if(get_nextToken(&to_read,fin)==OPEN_PAREN && get_nextToken(&to_read,fin)==TOTAL_COST_TOK && 
								get_nextToken(&to_read,fin)==CLOSE_PAREN){
					token = get_nextToken(&to_read,fin);
					int cost = atoi(yyltext);
					if(cost > gmax_action_cost)
						gmax_action_cost = cost;
					goperators[index]->cost = cost;
				}
#if(DEBUG)
				printf("op %d : %s [%d]\n",index,goperators[index]->name,goperators[index]->cost);fflush(stdout);
#endif
				break;
			}	
		}
	}
#if(DEBUG)
	printf("done\n");fflush(stdout);
#endif
	fclose(fin);

}


void scan_fct_file(char *fct_file)
{
	//my_m is our main data structure that will be used throughout the execution
	unsigned long i= 0;
  
  //line is a buffer for reading input
	line[MAX_LINE_LEN-1] = '\n';
	//char** in;
	char c;
	int token,j;
	int length;
	FILE *fin;

	if ( (fin = fopen( fct_file, "r" )) == NULL ) {
        	printf("\n\nEXIT: can't open % file.\n\n",fct_file);
 		exit(1);
    }
#if(DEBUG)
	printf("begin to scan problem file... ");
#endif
  	//main parsing loop
	while((c=getc(fin))!=EOF){
		//if(isspace(c)) continue; //skip any space at the beginning of the line
		ungetc(c,fin);

		read_line(line,fin);
		to_read = line;
		//in = &to_read;

		while( (*to_read) != '\n' ){
			token = get_Token(&to_read);
			if(token==COMMENT_TOK)
				break;
			if(token==TOTAL_COST_TOK){
				//get name of action
				if( get_nextToken(&to_read,fin)==CLOSE_PAREN && get_nextToken(&to_read,fin)!=CLOSE_PAREN ){
					ginit_cost = atoi(yyltext);
#if(DEBUG)
					printf("initialize total-cost to %d",atoi(yyltext));
#endif
				}	

			}
			break;
		}
	}
#if(DEBUG)
	printf("done!\n");
#endif
	fclose(fin);

}
